package cz.cvut.kbss.reporting.persistence.dao;

import cz.cvut.kbss.reporting.environment.Generator;
import cz.cvut.kbss.reporting.model.OccurrenceReport;
import cz.cvut.kbss.reporting.model.Person;
import cz.cvut.kbss.reporting.model.Tag;
import cz.cvut.kbss.reporting.persistence.BaseDaoTestRunner;
import java.time.LocalDateTime;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class OccurrenceReportDaoTest extends BaseDaoTestRunner {

    @Autowired
    private OccurrenceReportDao dao;

    @Autowired
    private PersonDao personDao;

    @Autowired
    private TagDao tagDao;

    private Person author;

    @Before
    public void setUp() {
        this.author = Generator.generatePerson();
        personDao.persist(author);
    }

    @Test
    public void findAllReturnsReportsOrderedByOccurrenceStartDescending() {
        final List<OccurrenceReport> reports = generateValidReports();
        Collections.shuffle(reports);
        dao.persist(reports);

        final List<OccurrenceReport> result = dao.findAll();
        assertEquals(reports.size(), result.size());
        assertOccurrenceStartDescendingOrder(result);
    }

    @Test
    public void findAllWithTagReturnsReportsOrderedByOccurrenceStartDescending() {
        final String searchedTagName = "searched-tag";
        final List<OccurrenceReport> reports = generateValidReports();

        tagDao.persist(generateValidTag(searchedTagName));
        tagDao.persist(generateValidTags());

        List<Tag> otherTags = tagDao.findAll();
        Tag searchedTag = otherTags.stream()
            .filter(t -> t.getName().equals(searchedTagName))
            .findFirst().get();
        otherTags.remove(searchedTag);


        populateRandomlyReportsWithTags(reports, otherTags);
        populateRandomlyReportsWithTags(
            reports,
            Collections.singletonList(searchedTag)
        );

        Collections.shuffle(reports);
        dao.persist(reports);
        assertOccurrenceStartDescendingOrder(dao.findAll(searchedTag));
    }


    @Test
    public void findAllWithTagReturnsOnlyRelevantReports() {
        final String searchedTagName = "searched-tag";
        final List<OccurrenceReport> reports = generateValidReports();

        tagDao.persist(generateValidTag(searchedTagName));
        tagDao.persist(generateValidTags());

        List<Tag> otherTags = tagDao.findAll();
        Tag searchedTag = otherTags.stream()
            .filter(t -> t.getName().equals(searchedTagName))
            .findFirst().get();
        otherTags.remove(searchedTag);


        populateRandomlyReportsWithTags(reports, otherTags);
        int populatedReports = populateRandomlyReportsWithTags(
            reports,
            Collections.singletonList(searchedTag)
        );

        dao.persist(reports);
        List<OccurrenceReport> foundReports = dao.findAll(searchedTag);
        long matchingReportsCount = foundReports.stream()
            .filter(r -> r.getTags().contains(searchedTag))
            .count();
        assertEquals(populatedReports, (int) matchingReportsCount);


    }

    private List<OccurrenceReport> generateReports() {
        return IntStream.range(0, Generator.randomInt(5, 10))
            .mapToObj(i -> Generator.generateReport())
            .collect(
                Collectors.toList());
    }

    private List<OccurrenceReport> generateValidReports() {
        List<OccurrenceReport> reports = generateReports();
        reports.forEach(r -> {
            r.setAuthor(author);
            r.setCreated(LocalDateTime.now());
        });
        return reports;
    }

    private List<Tag> generateTags() {
        return IntStream.range(0, Generator.randomInt(0, 5))
            .mapToObj(i -> Generator.generateTag("tag-" + Generator.randomInt()))
            .collect(Collectors.toList());
    }

    private List<Tag> generateValidTags() {
        List<Tag> tags = generateTags();
        tags.forEach(r -> {
            r.setAuthor(author);
            r.setCreated(LocalDateTime.now());
        });
        return tags;
    }


    private void assertOccurrenceStartDescendingOrder(List<OccurrenceReport> reports) {
        if (reports.size() == 0) {
            return;
        }

        OccurrenceReport previous = reports.get(0);
        for (int i = 1; i < reports.size(); i++) {
            final OccurrenceReport current = reports.get(i);
            assertTrue(current.getOccurrence().getTime().getStartTime().isBefore(previous.getOccurrence().getTime().getStartTime()));
            previous = current;
        }
    }

    private int populateRandomlyReportsWithTags(List<OccurrenceReport> reports, List<Tag> tags) {

        int populatedReports = 0;
        for (OccurrenceReport r : reports) {
            List<Tag> tagsSubset = Generator.getRandomSubset(tags);
            if (tagsSubset.isEmpty()) {
                break;
            }
            r.setTags(tagsSubset);
            populatedReports++;
        }
        return populatedReports;
    }


    private Tag generateValidTag(String searchedTagName) {
        Tag tag = Generator.generateTag(searchedTagName);
        tag.setAuthor(author);
        return tag;
    }
}