package cz.cvut.kbss.reporting.rest;

import cz.cvut.kbss.reporting.dto.PersonDto;
import cz.cvut.kbss.reporting.environment.Environment;
import cz.cvut.kbss.reporting.environment.Generator;
import cz.cvut.kbss.reporting.model.Person;
import cz.cvut.kbss.reporting.rest.mapper.DtoMapper;
import cz.cvut.kbss.reporting.rest.mapper.DtoMapperImpl;
import cz.cvut.kbss.reporting.service.repository.PersonService;
import cz.cvut.kbss.reporting.service.security.SecurityUtils;
import org.junit.Before;
import org.junit.Test;
import org.mockito.*;
import org.springframework.http.MediaType;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import java.util.Collections;

import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.header;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

public class PersonControllerTest extends BaseControllerTestRunner {

    @Mock
    private SecurityUtils securityUtilsMock;

    @Mock
    private PersonService personService;

    @Spy
    private DtoMapper dtoMapper = new DtoMapperImpl();

    @InjectMocks
    private PersonController controller;

    @Before
    public void setUp() {
        MockitoAnnotations.initMocks(this);
        super.setUp();
        this.mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
    }

    @Test
    public void getCurrentReturnsCurrentlyLoggedInUser() throws Exception {
        final Person person = Generator.generatePerson();
        person.setId(Generator.randomInt());
        Environment.setCurrentUser(person);
        when(securityUtilsMock.getCurrentUser()).thenReturn(person);
        MvcResult result = mockMvc.perform(get("/persons/current").principal(Environment.getCurrentUserPrincipal()))
                                  .andReturn();
        final Person res = readValue(result, Person.class);
        assertNotNull(res);
        assertEquals(person.getId(), res.getId());
        assertEquals(person.getUsername(), res.getUsername());
    }

    @Test
    public void createPersonPersistsNewPersonAndReturnsLocationHeader() throws Exception {
        authenticateAnonymously();
        final Person p = Generator.generatePerson();
        p.setId(Generator.randomInt());
        mockMvc.perform(post("/persons").content(toJson(p)).contentType(MediaType.APPLICATION_JSON))
               .andExpect(status().isCreated())
               .andExpect(header().string("location", containsString("/persons/" + p.getId())));
        final ArgumentCaptor<Person> captor = ArgumentCaptor.forClass(Person.class);
        verify(personService).persist(captor.capture());
        assertEquals(p.getUsername(), captor.getValue().getUsername());
    }

    private void authenticateAnonymously() {
        SecurityContext ctx = SecurityContextHolder.createEmptyContext();
        SecurityContextHolder.setContext(ctx);
        ctx.setAuthentication(new UsernamePasswordAuthenticationToken("anonymous", "",
                Collections.singletonList(new SimpleGrantedAuthority("ROLE_ANONYMOUS"))));
    }

    @Test
    public void updateUserUpdatedUserData() throws Exception {
        final Person person = Generator.generatePerson();
        person.setId(Generator.randomInt());
        Environment.setCurrentUser(person);
        final Person update = new Person();
        update.setId(person.getId());
        update.setFirstName("UpdatedFirstName");
        update.setLastName("UpdatedLastName");
        update.setUsername(person.getUsername());
        when(securityUtilsMock.getCurrentUser()).thenReturn(person);
        mockMvc.perform(put("/persons/current").content(toJson(update)).contentType(MediaType.APPLICATION_JSON)
                                               .principal(Environment.getCurrentUserPrincipal()))
               .andExpect(status().isNoContent());
        verify(personService).update(update);
        verify(securityUtilsMock, never()).verifyCurrentUserPassword(any());
    }

    @Test
    public void updateUserVerifiesOldPasswordWhenNewOneIsSpecified() throws Exception {
        final Person person = Generator.generatePerson();
        person.setId(Generator.randomInt());
        Environment.setCurrentUser(person);
        final TestPersonDto update = new TestPersonDto();
        update.setId(person.getId());
        update.setFirstName(person.getFirstName());
        update.setLastName(person.getLastName());
        update.setUsername(person.getUsername());
        update.setPassword("newPassword");
        update.setPasswordOriginal(person.getPassword());
        when(securityUtilsMock.getCurrentUser()).thenReturn(person);
        mockMvc.perform(put("/persons/current").content(toJson(update)).contentType(MediaType.APPLICATION_JSON)
                                               .principal(Environment.getCurrentUserPrincipal()))
               .andExpect(status().isNoContent());
        verify(securityUtilsMock).verifyCurrentUserPassword(update.getPasswordOriginal());
    }

    @Test
    public void doesUsernameExistReturnsValueForQuery() throws Exception {
        when(personService.exists(anyString())).thenReturn(true);
        final String username = "masterchief";
        final MvcResult mvcResult = mockMvc
                .perform(get("/persons/exists").param("username", username).accept(MediaType.TEXT_PLAIN_VALUE))
                .andExpect(status().isOk()).andReturn();
        final String result = mvcResult.getResponse().getContentAsString();
        assertTrue(Boolean.parseBoolean(result));
        verify(personService).exists(username);
    }

    private static class TestPersonDto extends PersonDto {
        // We are using this to bypass the WRITE_ONLY access to the password property in Person
        private String password;

        @Override
        public String getPassword() {
            return password;
        }

        @Override
        public void setPassword(String password) {
            this.password = password;
        }
    }
}